import sys
import os
import glob
import re
import common
forge = sys.modules["forge"]

def setup(module):
	manifest = 'ext/alembic/manifest.cc'
	with open(manifest, 'w') as outfile:
		outfile.write("")

	compiler = ''
	compilers_allowed = [ "gcc48" ]
	for check_version in compilers_allowed:
		if check_version in forge.compilers:
			compiler = forge.compilers[check_version]
			break

	for alembic_source in reversed(sorted(forge.alembic_sources)):
		alembic_source_version = alembic_source[0]
#		alembic_source_suffix = alembic_source[1]
		alembic_source_inc = alembic_source[2]
		alembic_source_libs = alembic_source[3]

		if not os.path.exists(alembic_source_inc) or not os.path.exists(alembic_source_libs):
			forge.color_on(0, forge.BLUE)
			sys.stdout.write(" -" + alembic_source_version)
			forge.color_off()
			continue

		alembic_static = False
		alembic_lib = common.find_alembic_lib(alembic_source_libs)
		if alembic_lib == '':
			alembic_static = True
			alembic_lib = common.find_alembic_static(alembic_source_libs)
			if alembic_lib == '':
				forge.color_on(0, forge.BLUE)
				sys.stdout.write(" -" + alembic_source_version)
				forge.color_off()
				continue

		if alembic_source_version == "":
			sys.stdout.write(" /")
		else:
			sys.stdout.write(" ")
			if alembic_static:
				sys.stdout.write("~")
			sys.stdout.write(alembic_source_version)

#		sys.stdout.write(" " + alembic_lib)

		srcList = [	"SurfaceAccessibleAbc",
					"alembic.pmh",
					"alembicDL" ]

		variant = alembic_source_version
		if variant != "":
			variantPath = module.modPath + '/' + variant
			if os.path.lexists(variantPath) == 0:
				os.symlink('.', variantPath)

			srcListVariant = []
			for src in srcList:
				srcListVariant += [ variant + '/' + src ]
			srcList = srcListVariant

		alembic_version = variant
		alembic_version_path = alembic_source_inc + '/Alembic/AbcCoreAbstract/Foundation.h'
		if os.path.exists(alembic_version_path):
			for line in open(alembic_version_path):
				if 'ALEMBIC_LIBRARY_VERSION ' in line:
					alembic_version = line.split(' ')[2].rstrip()
					break

		if alembic_version != variant:
			sys.stdout.write("=" + alembic_version)

		dllname = "fexAlembicDL" + variant

		with open(manifest, 'a') as outfile:
			suffix = ""
			if variant != "":
				suffix = "." + variant
			outfile.write('\tspManifest->catalog<String>(\n'+
					'\t\t\t"SurfaceAccessibleI.SurfaceAccessibleAbc.fe.abc'+
					suffix+'")=\n'+
					'\t\t\t"fexAlembicDL'+variant+'";\n');

		dll = module.DLL( dllname, srcList )

		for src in srcList:
			relativePath = forge.RelativePath(forge.rootPath,module.modPath)
			modObjPath = os.path.join(forge.objPath, relativePath)
			objName = modObjPath + '/' + src
			if re.compile(r'(.*)\.pmh$').match(src):
				objName += '.gch'
			else:
				objName += forge.objsuf
			srcTarget = forge.FindTarget(objName)
			srcTarget.includemap = { 'alembic' : alembic_source_inc }

			alembic_openexr = alembic_source_inc + '/OpenEXR'
			if os.path.exists(alembic_openexr):
				srcTarget.includemap['openexr'] = alembic_openexr
			else:
				alembic_openexr = alembic_source_inc + '/AlembicPrivate/OpenEXR'
				if os.path.exists(alembic_openexr):
					srcTarget.includemap['openexr'] = alembic_openexr
				elif forge.openexr_source != '':
					default_openexr = forge.openexr_source + '/include/OpenEXR'
					if os.path.exists(default_openexr):
						srcTarget.includemap['openexr'] = default_openexr

			if compiler != "":
				srcTarget.cxx = compiler

			srcTarget.cppmap = {}
			srcTarget.cppmap['std'] = '-std=c++11'

#~		dll.linkmap = { "alembic" :
#~				"-L" + alembic_source_libs + ' -l' + alembic_lib }
#~		if alembic_source_version == '':
#~			dll.linkmap["alembic"] += ' -Xlinker -rpath ' + alembic_source_libs

		alembicLibPath = os.path.join(forge.libPath, 'alembic' + variant)
		if os.path.isdir(alembicLibPath) == 0:
			os.makedirs(alembicLibPath, 0755)

		dll.linkmap = { "alembic" : "" }

		dll.linkmap["alembic"] += " -Xlinker -rpath '" + alembicLibPath + "/'"

		dll.linkmap["alembic"] += " -L" + alembic_source_libs + ' -l' + alembic_lib

		alembicPatterns = [
						"hdf5",
						"jemalloc",
						"Alembic",
						"Half",
						"Iex",
						"Imath" ]

		alembicLibs = [	]

		for alembicPattern in alembicPatterns:
			liblist = glob.glob(alembic_source_libs + '/lib' + alembicPattern + '*')
			for lib in liblist:
				alembicLib = os.path.basename(lib)
				alembicLibs += [ alembicLib ]

#		dll.linkmap['alembic'] += ' -lhdf5'
		dll.linkmap['alembic'] += ' -lAlembicAbc'
#		dll.linkmap['alembic'] += ' -lAlembicAbcCollection'
		dll.linkmap['alembic'] += ' -lAlembicAbcCoreAbstract'
		dll.linkmap['alembic'] += ' -lAlembicAbcCoreHDF5'
		dll.linkmap['alembic'] += ' -lAlembicAbcGeom'
#		dll.linkmap['alembic'] += ' -lAlembicAbcMaterial'

		if int(alembic_version) >= 10500:
			dll.linkmap['alembic'] += ' -lAlembicAbcCoreFactory'

		if os.path.exists(alembic_source_libs + "/libAlembicOgawa.so"):
			dll.linkmap['alembic'] += ' -lAlembicAbcCoreOgawa'
			dll.linkmap['alembic'] += ' -lAlembicOgawa'
			dll.linkmap['alembic'] += ' -lAlembicUtil'

		if os.path.exists(alembic_source_libs + "/libAlembicAbcOpenGL.so"):
			dll.linkmap['alembic'] += ' -lAlembicAbcOpenGL'

		if os.path.exists(alembic_source_libs + "/libHalf.so"):
			dll.linkmap['alembic'] += ' -lHalf'

		if os.path.exists(alembic_source_libs + "/libIex.so"):
			dll.linkmap['alembic'] += ' -lIex'

#!		alembicDsos = os.listdir(alembic_source_libs)
#!
#!		# NOTE Houdini dso paths have lots of other stuff (hundreds of files)
#!		if len(alembicDsos) < 20:
#!			sys.stdout.write("&")
#!
#!			# WARNING any rpath can bring in undesired alternatives of other libraries
#!			dll.linkmap["alembic"] += " -Xlinker -rpath '" + alembic_source_libs + "/'"

		for alembicLib in alembicLibs:
			src = os.path.join(alembic_source_libs, alembicLib)
			dest = os.path.join(alembicLibPath, alembicLib)
			if os.path.lexists(dest) == 0:
				os.symlink(src, dest)

		deplibs = forge.corelibs + [
					"fexSurfaceDLLib" ]

		forge.deps( [ dllname + "Lib" ], deplibs )

	forge.tests += [
		("inspect.exe",		"fexAlembicDL",				None,		None) ]

	module.Module('test')
