/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "window/window.h"
#include "draw/draw.h"

using namespace fe;


int main(int argc,char** argv)
{
	UNIT_START();

	BWORD complete=FALSE;

	try
	{
		sp<Master> spMaster(new Master);
		sp<Registry> spRegistry=spMaster->registry();

		Result result=spRegistry->manage("fexNativeWindowDL");
		UNIT_TEST(successful(result));

		result=spRegistry->manage("fexOpenGLDL");
		UNIT_TEST(successful(result));

		{
			sp<WindowI> spWindowI(spRegistry->create("WindowI"));
			sp<DrawI> spDrawI(spRegistry->create("DrawI"));

			if(!spWindowI.isValid() || !spDrawI.isValid())
			{
				feX(argv[0], "can't continue");
			}

			sp<EventContextI> spEventContextI=spWindowI->getEventContextI();
			if(!spEventContextI.isValid())
			{
				feX(argv[0], "can't continue");
			}

			UNIT_TEST(spEventContextI->fontBase());
		}

		complete=TRUE;
	}
	catch(Exception &e)
	{
		e.log();
	}
	catch(...)
	{
		feLog("uncaught exception\n");
	}

	UNIT_TEST(complete);
	UNIT_TEST(!Counted::trackerCount());
	UNIT_TRACK(5);
	UNIT_RETURN();
}

