/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "opencl/opencl.pmh"

#define FE_CCL_DEBUG	FALSE

namespace fe
{

CacheCLOp::CacheCLOp(void)
{
#if FE_CCL_DEBUG
	feLog("CacheCLOp::CacheCLOp\n");
#endif
}

CacheCLOp::~CacheCLOp(void)
{
#if FE_CCL_DEBUG
	feLog("CacheCLOp::~CacheCLOp\n");
#endif
}

void CacheCLOp::initialize(void)
{
#if FE_CCL_DEBUG
	feLog("CacheCLOp::initialize\n");
#endif

	catalog<String>("icon")="FE_alpha";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","clobber")=true;
}

void CacheCLOp::handle(Record& a_rSignal)
{
#if FE_CCL_DEBUG
	feLog("CacheCLOp::handle\n");
#endif

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

#if FE_CCL_DEBUG
	feLog("CacheCLOp::handle output %p cache %p\n",
			sp<SurfaceAccessibleOpenCL>(spOutputAccessible).raw(),
			sp<SurfaceAccessibleOpenCL>(m_spSurfaceAccessibleI).raw());

	const I32 pointCount=
			spOutputAccessible->count(SurfaceAccessibleI::e_point);
	const I32 primitiveCount=
			spOutputAccessible->count(SurfaceAccessibleI::e_primitive);
	feLog("  pointCount %d primitiveCount %d\n",
			pointCount,primitiveCount);
#endif

	sp<SurfaceAccessibleOpenCL> spOutputOpenCL(spOutputAccessible);
	if(spOutputOpenCL.isNull())
	{
		feLog("CacheCLOp::handle not OpenCL input\n");
		return;
	}

	if(m_spSurfaceAccessibleI.isNull())
	{
		feLog("CREATE CACHE\n");
		m_spSurfaceAccessibleI=sp<Protectable>(spOutputOpenCL->clone());
	}
	else
	{
		feLog("COPY CACHE\n");
		sp<SurfaceAccessibleOpenCL> spCacheOpenCL(m_spSurfaceAccessibleI);
		if(spCacheOpenCL.isNull())
		{
			feLog("CacheCLOp::handle not OpenCL cache\n");
			return;
		}

		spCacheOpenCL->clone(spOutputOpenCL.raw());
	}

#if FE_CCL_DEBUG
	feLog("CacheCLOp::handle done\n");
#endif
}

} // namespace
