/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "opencl/opencl.pmh"

#define FE_SAOC_CLONE_DEBUG		FALSE

namespace fe
{

sp<SurfaceAccessibleOpenCL::ClMem>
		SurfaceAccessibleOpenCL::Bridge::getClMem(void)
{
	safeLock();

//	feLog("SurfaceAccessibleOpenCL::Bridge::getClMem %d %p context %p\n",
//			m_spClMem.isValid(),
//			m_spClMem.isValid()? m_spClMem->m_clMem: nullptr,
//			m_clContext);

	sp<ClMem> spClMem=m_spClMem;

	if(spClMem.isNull() && m_clContext)
	{
		spClMem=new ClMem();

		if(spClMem.isValid())
		{
			//* TODO other types and rates
			cl_int bufferSize=m_count*m_sizeof;

			cl_int clError=0;
			spClMem->m_clMem=clCreateBuffer(m_clContext,
					CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR,
					bufferSize,m_bufferData,&clError);

			feLog("SurfaceAccessibleOpenCL::Bridge::getClMem"
					" clCreateBuffer %p clError %d\n",
					spClMem->m_clMem,clError);
			feLog("  \"%s\" \"%s\" \"%s\" count %d sizeof %d\n",
					m_name.c_str(),m_type.c_str(),m_rate.c_str(),
					m_count,m_sizeof);

			if(clError || spClMem->m_clMem==nullptr)
			{
				spClMem=NULL;
			}
		}

		m_spClMem=spClMem;
	}

	safeUnlock();

	return spClMem;
}

SurfaceAccessibleOpenCL::SurfaceAccessibleOpenCL(void):
	m_clContext(nullptr)
{
}

SurfaceAccessibleOpenCL::~SurfaceAccessibleOpenCL(void)
{
}

void SurfaceAccessibleOpenCL::reset(void)
{
}

Protectable* SurfaceAccessibleOpenCL::clone(Protectable* pInstance)
{
	SurfaceAccessibleOpenCL* pSurfaceAccessibleOpenCL= pInstance?
			dynamic_cast<SurfaceAccessibleOpenCL*>(pInstance):
			new SurfaceAccessibleOpenCL();

#if FE_SAOC_CLONE_DEBUG
	feLog("SurfaceAccessibleOpenCL::clone as %p from %p\n",
			pSurfaceAccessibleOpenCL,pInstance);
#endif

	updateContext();

	if(pSurfaceAccessibleOpenCL->library().isNull())
	{
		pSurfaceAccessibleOpenCL->setLibrary(library());
	}
	pSurfaceAccessibleOpenCL->m_clContext=m_clContext;

	for(std::map<String, sp<Bridge> >::iterator it=m_bufferMap.begin();
			it!=m_bufferMap.end();it++)
	{
		const String& bridgeName=it->first;
		const sp<Bridge>& rspSourceBridge=it->second;

#if FE_SAOC_CLONE_DEBUG
		feLog("  Bridge \"%s\"\n",bridgeName.c_str());
#endif

		if(rspSourceBridge.isValid())
		{
			sp<Bridge> spDestBridge=
					pSurfaceAccessibleOpenCL->lookupBridge(bridgeName,TRUE);
			spDestBridge->m_type=rspSourceBridge->m_type;
			spDestBridge->m_rate=rspSourceBridge->m_rate;
			spDestBridge->m_count=rspSourceBridge->m_count;
			spDestBridge->m_sizeof=rspSourceBridge->m_sizeof;
			spDestBridge->m_cpuCurrent=rspSourceBridge->m_cpuCurrent;
			spDestBridge->m_gpuCurrent=rspSourceBridge->m_gpuCurrent;

			//* TODO shouldn't ever be zero
			const size_t entrySize=
					spDestBridge->m_sizeof? spDestBridge->m_sizeof: sizeof(I32);

			const cl_int bufferSize=spDestBridge->m_count*entrySize;

			//* TODO check m_cpuCurrent and m_gpuCurrent

			if(rspSourceBridge->m_spClMem.isNull())
			{
				rspSourceBridge->getClMem();
			}

			if(rspSourceBridge->m_spClMem.isValid())
			{
				cl_mem clSourceMem=rspSourceBridge->m_spClMem->m_clMem;
				if(clSourceMem)
				{
					sp<ClMem> spNewClMem(new ClMem());

					if(spNewClMem.isValid())
					{

						cl_int clError=0;
						spNewClMem->m_clMem=clCreateBuffer(m_clContext,
								CL_MEM_READ_WRITE,
								bufferSize,nullptr,&clError);

#if FE_SAOC_CLONE_DEBUG
						feLog("SurfaceAccessibleOpenCL::clone"
								" clCreateBuffer %p clError %d\n",
								spNewClMem->m_clMem,clError);
#endif

						if(clError || spNewClMem->m_clMem==nullptr)
						{
							spNewClMem=NULL;
						}
						else
						{
							sp<Catalog> spMasterCatalog=
									registry()->master()->catalog();
							cl_command_queue clCommandQueue=(cl_command_queue)
									spMasterCatalog->catalogOrDefault<void*>(
									"clCommandQueue",NULL);

							clError=clEnqueueCopyBuffer(
									clCommandQueue,clSourceMem,
									spNewClMem->m_clMem,
									0,0,bufferSize,
									0,NULL,NULL);
#if FE_SAOC_CLONE_DEBUG
							feLog("clEnqueueCopyBuffer %p to %p"
									" queue %p clError %d\n",
									clSourceMem,spNewClMem->m_clMem,
									clCommandQueue,clError);
#endif

							clError=clFinish(clCommandQueue);
#if FE_SAOC_CLONE_DEBUG
							feLog("clFinish clError %d\n",clError);
#endif
						}
					}

					spDestBridge->safeLock();
					spDestBridge->m_spClMemPrevious=spDestBridge->m_spClMem;
					spDestBridge->m_spClMem=spNewClMem;
					spDestBridge->safeUnlock();
				}
			}
			else if(rspSourceBridge->m_bufferData)
			{
				spDestBridge->m_allocated=rspSourceBridge->m_allocated;
				const size_t allocSize=spDestBridge->m_allocated*entrySize;

#if FE_SAOC_CLONE_DEBUG
				feLog("    copy bufferData %d %d\n",bufferSize,allocSize);
#endif
				spDestBridge->m_bufferData=reallocate(nullptr,allocSize);

				memcpy(spDestBridge->m_bufferData,
						rspSourceBridge->m_bufferData,bufferSize);
			}
		}
	}

#if FE_SAOC_CLONE_DEBUG
	feLog("SurfaceAccessibleOpenCL::clone result %p\n",
			pSurfaceAccessibleOpenCL);
#endif

	return pSurfaceAccessibleOpenCL;
}

void SurfaceAccessibleOpenCL::attributeSpecs(
	Array<SurfaceAccessibleI::Spec>& a_rSpecs,
	SurfaceAccessibleI::Element a_element) const
{
//	feLog("SurfaceAccessibleOpenCL::attributeSpecs element %d\n",a_element);

	a_rSpecs.clear();

	//* TODO

	SurfaceAccessibleI::Spec spec;

	switch(a_element)
	{
		case SurfaceAccessibleI::e_point:
			spec.set("P","vector3");
			a_rSpecs.push_back(spec);
			break;
		case SurfaceAccessibleI::e_vertex:
			break;
		case SurfaceAccessibleI::e_primitive:
			break;
		case SurfaceAccessibleI::e_detail:
			break;
		default:
			return;
	}
}

sp<SurfaceAccessorI> SurfaceAccessibleOpenCL::accessor(
	String a_node,SurfaceAccessibleI::Element a_element,
	String a_name,SurfaceAccessibleI::Creation a_create)
{
	sp<SurfaceAccessorOpenCL> spAccessor=createAccessor();
	if(spAccessor.isValid())
	{
		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
		if(spAccessor->bind(a_element,a_name) ||
				a_create==SurfaceAccessibleI::e_createMissing)
		{
			return spAccessor;
		}
	}
	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorI> SurfaceAccessibleOpenCL::accessor(
	String a_node,SurfaceAccessibleI::Element a_element,
	SurfaceAccessibleI::Attribute a_attribute,
	SurfaceAccessibleI::Creation a_create)
{
	sp<SurfaceAccessorOpenCL> spAccessor=createAccessor();
	if(spAccessor.isValid())
	{
		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
		if(spAccessor->bind(a_element,a_attribute) ||
				a_create==SurfaceAccessibleI::e_createMissing)
		{
			return spAccessor;
		}
	}
	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorI> SurfaceAccessibleOpenCL::createAccessor(void)
{
	sp<SurfaceAccessorOpenCL> spAccessorOpenCL(new SurfaceAccessorOpenCL());

	return spAccessorOpenCL;
}

BWORD SurfaceAccessibleOpenCL::discard(
	SurfaceAccessibleI::Element a_element,String a_name)
{
	//* TODO
	return FALSE;
}

void SurfaceAccessibleOpenCL::updateContext(void)
{
	if(!m_clContext)
	{
		sp<Catalog> spMasterCatalog=registry()->master()->catalog();
		m_clContext=(cl_context)spMasterCatalog->catalogOrDefault<void*>(
				"clContext",nullptr);

		if(m_clContext)
		{
			for(std::map<String, sp<Bridge> >::iterator it=m_bufferMap.begin();
					it!=m_bufferMap.end();it++)
			{
				sp<Bridge>& rspBridge=it->second;
				if(rspBridge.isValid())
				{
					rspBridge->m_clContext=m_clContext;
				}
			}
		}
	}
}

sp<SurfaceAccessibleOpenCL::Bridge> SurfaceAccessibleOpenCL::lookupBridge(
	String a_name,BWORD a_create)
{
	updateContext();

	sp<Bridge> spBridge;

	if(a_create || m_bufferMap.find(a_name)!=m_bufferMap.end())
	{
		spBridge=m_bufferMap[a_name];
		if(spBridge.isNull())
		{
			if(a_create)
			{
				spBridge=sp<Bridge>(new Bridge(m_clContext));
				spBridge->m_name=a_name;

				m_bufferMap[a_name]=spBridge;
			}
			else
			{
				m_bufferMap.erase(a_name);
			}
		}
	}

	return spBridge;
}

sp<SurfaceI> SurfaceAccessibleOpenCL::surface(String a_group)
{
	feLog("SurfaceAccessibleOpenCL::surface(\"%s\")\n",a_group.c_str());

	//* TODO select form
	//* TODO group
	sp<SurfaceI> spSurface;

	FEASSERT(registry().isValid());

	if(registry().isValid())
	{
		sp<SurfaceAccessorI> spSurfacePoints=
				accessor(e_point,e_position,e_refuseMissing);
		const I32 pointCount=spSurfacePoints.isValid()?
				spSurfacePoints->count(): 0;

		sp<SurfaceAccessorI> spSurfaceVertices=
				accessor(e_primitive,e_vertices,e_refuseMissing);
		const I32 primitiveCount=spSurfaceVertices.isValid()?
				spSurfaceVertices->count(): 0;

		const BWORD pointCloud=(pointCount && !primitiveCount);

		sp<SurfaceAccessorI> spSurfaceProperties=
				accessor(e_primitive,e_properties,e_refuseMissing);

		const BWORD openCurve=(spSurfaceProperties.isValid() &&
				spSurfaceProperties->count() &&
				spSurfaceProperties->integer(0,
				SurfaceAccessibleI::e_openCurve));

		feLog("SurfaceAccessibleOpenCL::surface"
				" pointCount %d primitiveCount %d"
				" pointCloud %d openCurve %d\n",
				pointCount,primitiveCount,pointCloud,openCurve);

		if(pointCloud || openCurve)
		{
			feLog("CREATE SurfaceCurvesOpenCL\n");

			spSurface=registry()->create("*.SurfaceCurvesOpenCL");

			sp<SurfaceCurvesOpenCL> spSurfaceCurvesOpenCL=spSurface;
			if(spSurfaceCurvesOpenCL.isValid())
			{
				feLog("is SurfaceCurvesOpenCL\n");

				spSurfaceCurvesOpenCL->bind(
						sp<SurfaceAccessibleI>(this));
			}
		}
		else
		{
			feLog("CREATE SurfaceTrianglesOpenCL\n");

			spSurface=registry()->create(
					"*.SurfaceTrianglesOpenCL");

			sp<SurfaceTrianglesOpenCL> spSurfaceTrianglesOpenCL=spSurface;
			if(spSurfaceTrianglesOpenCL.isValid())
			{
				feLog("is SurfaceTrianglesOpenCL\n");

				spSurfaceTrianglesOpenCL->bind(
						sp<SurfaceAccessibleI>(this));
			}
		}
	}
	else
	{
		feLog("SurfaceAccessibleOpenCL::surface registry is NULL\n");
	}

	if(spSurface.isNull())
	{
		feLog("SurfaceAccessibleOpenCL::surface failed to access surface\n");
	}

	return spSurface;
}
} // namespace
