/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __opencl_SurfaceAccessibleOpenCL_h__
#define __opencl_SurfaceAccessibleOpenCL_h__

namespace fe
{

/**************************************************************************//**
    @brief OpenCL Surface Binding

	@ingroup opencl
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessibleOpenCL:
	public SurfaceAccessibleBase
{
	public:
								SurfaceAccessibleOpenCL(void);
virtual							~SurfaceAccessibleOpenCL(void);

								//* As Protectable
virtual	Protectable*			clone(Protectable* pInstance=NULL);

								//* as SurfaceAccessibleI
virtual	BWORD					isBound(void)
								{	return TRUE; }

								using SurfaceAccessibleBase::attributeSpecs;

virtual	void					attributeSpecs(
										Array<SurfaceAccessibleI::Spec>&
										a_rSpecs,
										SurfaceAccessibleI::Element
										a_element) const;

								using SurfaceAccessibleBase::accessor;

virtual sp<SurfaceAccessorI>	accessor(String a_node,
										SurfaceAccessibleI::Element a_element,
										String a_name,
										SurfaceAccessibleI::Creation a_create);
virtual sp<SurfaceAccessorI>	accessor(String a_node,
										SurfaceAccessibleI::Element a_element,
										SurfaceAccessibleI::Attribute
										a_attribute,
										SurfaceAccessibleI::Creation a_create);

								using SurfaceAccessibleBase::discard;

virtual	BWORD					discard(SurfaceAccessibleI::Element a_element,
										String a_name);

								using SurfaceAccessibleBase::surface;

virtual	sp<SurfaceI>			surface(String a_group);

								//* OpenCL specific

	class ClMem: public Counted
	{
		public:
					ClMem(void):
						m_clMem(NULL)										{}

					~ClMem(void)
					{	clReleaseMemObject(m_clMem); }

			cl_mem		m_clMem;
	};

	class Bridge:
		public ObjectSafeShared<Bridge>,
		public Counted
	{
		public:
					Bridge(cl_context a_clContext):
						m_bufferData(NULL),
						m_count(0),
						m_allocated(0),
						m_sizeof(0),
						m_clContext(a_clContext),
						m_cpuCurrent(TRUE),
						m_gpuCurrent(TRUE)									{}

					~Bridge(void)
					{	deallocate(m_bufferData); }

			sp<ClMem>	getClMem(void);

			String		m_name;
			String		m_type;
			String		m_rate;

			void*		m_bufferData;
			I32			m_count;
			I32			m_allocated;
			I32			m_sizeof;

			cl_context	m_clContext;
			sp<ClMem>	m_spClMem;
			sp<ClMem>	m_spClMemPrevious;

						//* TODO
			BWORD		m_cpuCurrent;
			BWORD		m_gpuCurrent;
	};

		sp<Bridge>				lookupBridge(String a_name,BWORD a_create);

	protected:
virtual	void					reset(void);

	private:
		void					updateContext(void);

		sp<SurfaceAccessorI>	createAccessor(void);

		std::map< String, sp<Bridge> >	m_bufferMap;

		cl_context						m_clContext;
};

} // namespace

#endif /* __opencl_SurfaceAccessibleOpenCL_h__ */
