/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#define	FE_TMD_DEBUG		FALSE
#define	FE_TMD_OPENCL		TRUE	//* TODO conditional

#include "terminal/terminal.pmh"

namespace fe
{

TerminalDispatch::TerminalDispatch(void)
{
#if	FE_TMD_DEBUG
	feLog("TerminalDispatch::TerminalDispatch\n");
#endif

#if FE_COUNTED_TRACK
	Counted::registerRegion(this,sizeof(TerminalDispatch),"TerminalDispatch");
#endif
}

TerminalDispatch::~TerminalDispatch(void)
{
#if	FE_TMD_DEBUG
	feLog("TerminalDispatch::~TerminalDispatch\n");
#endif

#if FE_COUNTED_TRACK
	Counted::deregisterRegion(this);
#endif
}

bool TerminalDispatch::call(const String &a_name,Array<Instance>& a_argv)
{
#if	FE_TMD_DEBUG
	feLog("TerminalDispatch::call ordered \"%s\" %d\n",
			a_name.c_str(),a_argv.size());
#endif

	//* NOTE a_argv[0] is for output

	const I32 argc=a_argv.size();

#if	FE_TMD_DEBUG
	for(I32 index=0;index<argc;index++)
	{
		feLog("  %d/%d \"%s\"\n",index,argc,
				a_argv[index].data()? c_print(a_argv[index]): "NULL");
	}
#endif

	InstanceMap instanceMap;

	sp<Catalog> spCatalog=m_terminalNode.catalog();

	if(spCatalog.isValid() && argc>0)
	{
		const String outputKey=spCatalog->catalog<String>("output:0");
		instanceMap[outputKey]=a_argv[0];

		for(I32 index=1;index<argc;index++)
		{
			Instance& rInstance=a_argv[index];

//~			sp<SurfaceAccessibleI> spInputAccessible;
//~
//~			if(rInstance.data())
//~			{
//~				spInputAccessible=registry()->create(
//~						"*.SurfaceAccessibleRecord");
//~
//~				sp<SurfaceAccessibleRecord> spSurfaceAccessibleRecord=
//~						spInputAccessible;
//~				if(spSurfaceAccessibleRecord.isValid())
//~				{
//~					spSurfaceAccessibleRecord->bind(rInstance.cast<Record>());
//~				}
//~			}

			const I32 channel=index-1;

			String lookup;
			lookup.sPrintf("input:%d",channel);

			const String inputKey=spCatalog->catalog<String>(lookup);
			FEASSERT(!inputKey.empty());

			instanceMap[inputKey]=rInstance;
		}
	}

	if(!call(a_name,instanceMap))
	{
		return false;
	}

	if(spCatalog.isValid())
	{
		const String outputKey=spCatalog->catalog<String>("output:0");
		a_argv[0]=instanceMap[outputKey];
	}

	return true;
}

bool TerminalDispatch::call(const String &a_name,InstanceMap& a_argv)
{
	const Real frame=a_argv.exists("frame")?
			Real(a_argv["frame"].cast<double>()): Real(0);

#if	FE_TMD_DEBUG
	feLog("TerminalDispatch::call map \"%s\" %d frame %.6G\n",
			a_name.c_str(),a_argv.size(),frame);
#endif

	sp<Catalog> spCatalog=m_terminalNode.catalog();

	const String outputKey=spCatalog.isValid()?
			spCatalog->catalog<String>("output:0"): String();

	//* TODO global 'allow clobber' variable

	BWORD recycle=FALSE;
	BWORD clobber=FALSE;
	String copy;

	if(spCatalog.isValid())
	{
		copy=spCatalog->catalogOrDefault<String>(outputKey,"copy","");
		clobber=spCatalog->catalogOrDefault<bool>(outputKey,"clobber",false);
		recycle=spCatalog->catalogOrDefault<bool>(outputKey,"recycle",false);

#if FE_TMD_DEBUG
		feLog("  output \"%s\" copies \"%s\" recycle %d\n",
				outputKey.c_str(),copy.c_str(),recycle);
#endif

		for(InstanceMap::iterator it=a_argv.begin();it!=a_argv.end();it++)
		{
			const String key=it->first;
			Instance& rInstance=it->second;

#if	FE_TMD_DEBUG
			feLog("  key \"%s\"\n",key.c_str());
#endif

			if(key==outputKey || !rInstance.is<Record>())
			{
				continue;
			}

			sp<SurfaceAccessibleI> spInputAccessible;

			if(rInstance.data())
			{
				Record record=rInstance.cast<Record>();

				if(m_accessorAccessible.scope()==record.layout()->scope() &&
						m_accessorAccessible.check(record))
				{
#if	FE_TMD_DEBUG
					feLog("  as Component\n");
#endif
					spInputAccessible=m_accessorAccessible(record);
				}
				else
				{
#if	FE_TMD_DEBUG
					feLog("  as SurfaceAccessibleRecord\n");
#endif
					spInputAccessible=registry()->create(
							"*.SurfaceAccessibleRecord");

					sp<SurfaceAccessibleRecord> spSurfaceAccessibleRecord=
							spInputAccessible;
					if(spSurfaceAccessibleRecord.isValid())
					{
						spSurfaceAccessibleRecord->bind(record);
					}
				}
			}

			spCatalog->catalog< sp<Component> >(key)=spInputAccessible;
		}
	}

	//* TODO respect recycle

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(spCatalog.isValid() && !copy.empty())
	{
		sp<SurfaceAccessibleI> spSourceAccessible=
				spCatalog->catalog< sp<Component> >(copy);

		if(clobber)
		{
#if	FE_TMD_DEBUG
			feLog("  CLOBBER INPUT \"%s\"\n",copy.c_str());
#endif
			spOutputAccessible=spSourceAccessible;
		}
		else
		{
#if	FE_TMD_DEBUG
			feLog("  COPY INPUT  \"%s\"\n",copy.c_str());
#endif
			sp<Protectable> spProtectable(spSourceAccessible->clone());
			spOutputAccessible=spProtectable;
		}

		FEASSERT(spOutputAccessible.isValid());

		spOutputAccessible->setName(outputKey);

		spCatalog->catalog< sp<Component> >(outputKey)=
				spOutputAccessible;
	}
	else
	{
#if	FE_TMD_DEBUG
			feLog("  CREATE NEW OUTPUT\n");
#endif
		spOutputAccessible=spCatalog->catalog< sp<Component> >(outputKey);
		if(spOutputAccessible.isNull())
		{
#if FE_TMD_OPENCL
			spOutputAccessible=registry()->create("*.SurfaceAccessibleOpenCL");
#else
			spOutputAccessible=registry()->create("*.SurfaceAccessibleRecord");
			if(spOutputAccessible.isValid())
			{
//~				sp<Scope> spScope=m_spScope;

				//* independent attributes per mesh
				sp<Scope> spScope=registry()->create("Scope");
				spScope->setLocking(FALSE);

				spOutputAccessible->bind(spScope);
				spOutputAccessible->clear();
			}
#endif

			if(spCatalog.isValid())
			{
				spCatalog->catalog< sp<Component> >(outputKey)=
						spOutputAccessible;
			}
		}

		if(spOutputAccessible.isValid())
		{
			spOutputAccessible->clear();
		}
	}

#if	FE_TMD_DEBUG
	feLog("  COOK\n");
#endif

	m_terminalNode.cook(frame);

#if	FE_TMD_DEBUG
	feLog("  COOKED points %d primitives %d\n",
			spOutputAccessible->count(SurfaceAccessibleI::e_point),
			spOutputAccessible->count(SurfaceAccessibleI::e_primitive));
#endif

	if(spCatalog.isValid() && a_argv.size()>0)
	{
		sp<SurfaceAccessibleRecord> spSurfaceAccessibleRecord=
				spOutputAccessible;
		if(spSurfaceAccessibleRecord.isValid())
		{
			//* keep reference to Record for persistance
			m_outputRecord=spSurfaceAccessibleRecord->record();
		}
		else
		{
			m_outputRecord=m_spScope->createRecord(m_spSurfaceLayout);
			m_accessorAccessible(m_outputRecord)=spOutputAccessible;
		}

		const sp<TypeMaster>& spTM = registry()->master()->typeMaster();
		a_argv[outputKey].set<Record>(spTM,m_outputRecord);
	}

#if	FE_TMD_DEBUG
	feLog("TerminalDispatch::call \"%s\" done\n",a_name.c_str());
#endif

	return true;
}

bool TerminalDispatch::configLookup(const String &a_key,Instance &a_instance)
{
#if	FE_TMD_DEBUG
	feLog("TerminalDispatch::configLookup \"%s\"\n",a_key.c_str());
#endif

	if(a_key=="_scope_")
	{
		const sp<TypeMaster>& spTM = registry()->master()->typeMaster();
		a_instance.create< sp<Scope> >(spTM);

#if	FE_TMD_DEBUG
		feLog("  token scope\n");
#endif

		return true;
	}


	sp<Catalog> spCatalog=m_terminalNode.catalog();
	if(spCatalog.isNull())
	{
#if	FE_TMD_DEBUG
		feLog("  no catalog\n");
#endif
		return false;
	}

	Instance instance;
	if(spCatalog->catalogLookup(a_key,a_instance))
	{
#if	FE_TMD_DEBUG
		feLog("  key not found\n");
#endif
		return false;
	}

#if	FE_TMD_DEBUG
	feLog("  catalog entry \"%s\"\n",c_print(a_instance));
#endif

	return true;
}

void TerminalDispatch::configSet(const String &a_key, Instance &a_instance)
{
#if	FE_TMD_DEBUG
	feLog("TerminalDispatch::configSet \"%s\" \"%s\"\n",
			a_key.c_str(),c_print(a_instance));
#endif

	if(a_key=="_scope_")
	{
		m_spScope=a_instance.cast< sp<Scope> >();

		m_spSurfaceLayout=m_spScope->declare("surface");
		m_spScope->support("accessible","Component");
		m_spScope->populate("surface","accessible");
		m_accessorAccessible.setup(m_spScope,"accessible");

		return;
	}

	if(a_key=="_implementation_")
	{
		if(m_spScope.isNull())
		{
			feLog("TerminalDispatch::configSet implementation without scope\n");
			return;
		}

		const String implementation=a_instance.cast<String>();
		const String name=implementation;

		m_terminalNode.setupOperator(m_spScope,implementation,name);

		return;
	}

	sp<Catalog> spCatalog=m_terminalNode.catalog();
	if(spCatalog.isNull())
	{
		feLog("  no catalog\n");
		return;
	}

	spCatalog->catalogInstance(a_key)=a_instance;
}

} // namespace
