/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_MTX_COUNT	(FE_CODEGEN<FE_DEBUG)

#ifndef __platform_mutex_h__
#define __platform_mutex_h__

#if 0
#include "boost/thread.hpp"
#define FE_YIELD boost::thread::yield
#define FE_SLEEP(s, ns) \
	struct boost::xtime boost_time; \
	boost_time.sec=seconds; \
	boost_time.nsec=nanoseconds; \
	boost::thread::sleep(boost_time);
#endif

#if 0
#include "boost/thread.hpp"
#include "boost/thread/thread.hpp"
#include "boost/thread/tss.hpp"
#include "boost/thread/condition.hpp"
typedef boost::shared_mutex mutex;
typedef boost::recursive_mutex recursive_mutex;
#define FE_YIELD boost::thread::yield
#define FE_SLEEP(s, ns) \
	struct boost::xtime boost_time; \
	boost_time.sec=seconds; \
	boost_time.nsec=nanoseconds; \
	boost::thread::sleep(boost_time);
#define FE_LOCK(s, m) s.lock()
#define FE_UNLOCK(s) s.unlock()
#define FE_M_LOCK(m) m->lock()
#define FE_M_UNLOCK(m) m->unlock()
#define FE_M_TRY_LOCK(m) m->try_lock()
#define FE_M_LOCK_SHARED(m) m->lock_shared()
#define FE_M_UNLOCK_SHARED(m) m->unlock_shared()
#define FE_M_TRY_LOCK_SHARED(m) m->try_lock_shared()
#endif

#if 0
#include "tbb/mutex.h"
#include "tbb/recursive_mutex.h"
#include "tbb/spin_rw_mutex.h"
#include "tbb/tick_count.h"
#include "tbb/tbb_thread.h"
typedef tbb::spin_rw_mutex mutex;
typedef tbb::recursive_mutex recursive_mutex;
#define FE_YIELD tbb::this_tbb_thread::yield
#define FE_SLEEP(s, ns) \
	tbb::tick_count::interval_t t(((double)s) + (((double)ns) * 1.0e-9)); \
	tbb::this_tbb_thread::sleep(t)
#define FE_LOCK(s, m) s.acquire(m)
#define FE_UNLOCK(s) s.release()
#define FE_M_LOCK(m) m->lock()
#define FE_M_UNLOCK(m) m->unlock()
#define FE_M_TRY_LOCK(m) m->try_lock()
#define FE_M_LOCK_SHARED(m) m->lock_read()
#define FE_M_UNLOCK_SHARED(m) m->unlock()
#define FE_M_TRY_LOCK_SHARED(m) m->try_lock_read()
#endif

namespace fe
{

extern "C"
{

typedef void* (MutexInitFunction)(bool a_recursive);
typedef bool (MutexLockFunction)(bool a_recursive,void* a_pRawMutex,
		bool a_un,bool a_try,bool a_readOnly);
typedef void (MutexFinishFunction)(bool a_recursive,void* a_pRawMutex);

typedef void* (MutexGuardInitFunction)(
		bool a_recursive,void* a_pRawMutex,bool a_readOnly);
typedef bool (MutexGuardLockFunction)(bool a_recursive,void* a_pRawGuard,
		bool a_un,bool a_try,bool a_readOnly);
typedef void (MutexGuardFinishFunction)(bool a_recursive,void* a_pRawGuard,
		bool a_readOnly);

typedef void* (MutexConditionInitFunction)(void);
typedef bool (MutexConditionWaitFunction)(void* a_pRawCondition,
		bool a_recursive,void* a_pRawGuard,bool a_readOnly);
typedef bool (MutexConditionNotifyFunction)(void* a_pRawCondition,bool a_all);
typedef void (MutexConditionFinishFunction)(void* a_pRawCondition);

} /* extern "C" */

/*
	boost::mutex			m_jobMonitor;
	boost::mutex::scoped_lock lock(m_jobMonitor);

	boost::condition		m_jobAvailable;
	m_jobAvailable.wait(lock);
	m_jobAvailable.notify_one();
	m_barrierHit.notify_all();
*/

class FE_DL_PUBLIC FE_DL_EXPORT Mutex
{
	public:
				Mutex(BWORD a_recursive=FALSE);
				~Mutex(void);

static	BWORD	supported(void)		{ return (ms_fnInit!=NULL); }
static	String	support(void)		{ return ms_support; }

		BWORD	lock(void);
		BWORD	unlock(void);
		BWORD	tryLock(void);
		BWORD	lockReadOnly(void);
		BWORD	unlockReadOnly(void);
		BWORD	tryLockReadOnly(void);

	class Condition;
	class Guard
	{
		friend class Mutex::Condition;

		public:
					Guard(Mutex& a_rMutex,BWORD a_readOnly=FALSE,
							BWORD a_truly=TRUE);
					~Guard(void);

			BWORD	lock(void);
			BWORD	unlock(void);
			BWORD	tryLock(void);

		private:
			Mutex*	m_pMutex;
			void*	m_pRawGuard;
			bool	m_readOnly;
	};

	class Condition
	{
		public:
					Condition(void);
					~Condition(void);

			BWORD	wait(Guard& a_rGuard);
			BWORD	notifyOne(void);
			BWORD	notifyAll(void);

		private:
			void*	raw(void);

			void*	m_pRawCondition;
	};

static	void	replaceInitFunction(MutexInitFunction* a_fnInit)
				{	ms_fnInit=a_fnInit; }
static	void	replaceLockFunction(MutexLockFunction* a_fnLock)
				{	ms_fnLock=a_fnLock; }
static	void	replaceFinishFunction(MutexFinishFunction* a_fnFinish)
				{	ms_fnFinish=a_fnFinish; }

static	void	replaceGuardInitFunction(MutexGuardInitFunction* a_fnGuardInit)
				{	ms_fnGuardInit=a_fnGuardInit; }
static	void	replaceGuardLockFunction(MutexGuardLockFunction* a_fnGuardLock)
				{	ms_fnGuardLock=a_fnGuardLock; }
static	void	replaceGuardFinishFunction(
						MutexGuardFinishFunction* a_fnGuardFinish)
				{	ms_fnGuardFinish=a_fnGuardFinish; }

static	void	replaceConditionInitFunction(
						MutexConditionInitFunction* a_fnConditionInit)
				{	ms_fnConditionInit=a_fnConditionInit; }
static	void	replaceConditionWaitFunction(
						MutexConditionWaitFunction* a_fnConditionWait)
				{	ms_fnConditionWait=a_fnConditionWait; }
static	void	replaceConditionNotifyFunction(
						MutexConditionNotifyFunction* a_fnConditionNotify)
				{	ms_fnConditionNotify=a_fnConditionNotify; }
static	void	replaceConditionFinishFunction(
						MutexConditionFinishFunction* a_fnConditionFinish)
				{	ms_fnConditionFinish=a_fnConditionFinish; }

static	BWORD	confirm(String a_hint="");

	protected:

		BWORD	m_recursive;

	private:

		void*	raw(void);

static	MutexInitFunction*				ms_fnInit;
static	MutexLockFunction*				ms_fnLock;
static	MutexFinishFunction*			ms_fnFinish;

static	MutexGuardInitFunction*			ms_fnGuardInit;
static	MutexGuardLockFunction*			ms_fnGuardLock;
static	MutexGuardFinishFunction*		ms_fnGuardFinish;

static	MutexConditionInitFunction*		ms_fnConditionInit;
static	MutexConditionWaitFunction*		ms_fnConditionWait;
static	MutexConditionNotifyFunction*	ms_fnConditionNotify;
static	MutexConditionFinishFunction*	ms_fnConditionFinish;

static	String							ms_support;
static	BWORD							ms_checked;

		bool	m_initializing;
		void*	m_pRawMutex;

#if FE_MTX_COUNT
		I32		m_fakeCount;
		I32		m_lockCount;
#endif
};

class FE_DL_PUBLIC FE_DL_EXPORT RecursiveMutex: public Mutex
{
	public:
				RecursiveMutex(void):
					Mutex(TRUE)												{}
};

} /* namespace */

#endif /* __platform_mutex_h__ */
