/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "platform.pmh"

namespace fe
{

FE_DL_PUBLIC BWORD System::ms_bannerPrinted=FALSE;

void System::printBannerOnce(void)
{
	if(ms_bannerPrinted)
	{
		return;
	}

	const String verbose=System::getVerbose();

	if(verbose!="none")
	{
		printf("Free Electron - %s\n",System::buildDate().c_str());

#if FE_CODEGEN==FE_VALIDATE
			printf("%susing validate build\n",PrefixLog::prefix().c_str());
#elif FE_CODEGEN==FE_DEBUG
			printf("%susing debug build\n",PrefixLog::prefix().c_str());
#elif FE_CODEGEN==FE_PROFILE
			printf("%susing profile build\n",PrefixLog::prefix().c_str());
#endif
	}

	ms_bannerPrinted=TRUE;
}

bool System::getEnvironmentVariable(const String &name, String &value)
{
	bool rv = false;
#if FE_OS==FE_LINUX || FE_OS==FE_OSX
	char *cp;
	cp = getenv(name.c_str());
	if ((char *) 0 != cp)
	{
		value = cp;
		rv = true;
	}
#elif FE_OS==FE_WIN32 || FE_OS==FE_WIN64
	unsigned long cp;
	const int size = 4096;
	char val[size];
	cp = GetEnvironmentVariable(name.c_str(), val, size);
	if (0 != cp)
	{
		value = val;
		rv = true;
	}
#else
#error
#endif
	return rv;
}

String System::getVerbose(void)
{
	String verbose;

#if FE_CODEGEN<=FE_DEBUG
	verbose="all";
#else
	System::getEnvironmentVariable("FE_VERBOSE", verbose);
	if(verbose.empty())
	{
		verbose="none";
	}
#endif

	return verbose;
}

//* static
String System::buildDate(void)
{
	return __DATE__;
}

//* static
String System::getLoadPath(void* a_symbol,BWORD a_prune)
{
#if FE_OS==FE_LINUX || FE_OS==FE_OSX
	Dl_info dlInfo;
	if(!dladdr(a_symbol,&dlInfo))
	{
		fprintf(stderr,"getLoadPath failed\n");
		return "";
	}

	String path=dlInfo.dli_fname;

	if(!a_prune)
	{
		return path;
	}

	// NOTE Boost can miss the second /../ of /../../ since the patterns overlap
	// so we repeat the operator

	String oldPath;
	while(path!=oldPath)
	{
		oldPath=path;
		path=path.replace("/\\.\\./","/@@/");
	}

	oldPath="";
	while(path!=oldPath)
	{
		oldPath=path;
		path=oldPath.replace("/[^/@]*/@@/","/");
	}

#if FALSE
	path=path.replace("/[^/]*$","");
	path=path.replace("/lib/.*$","");
	path=path.replace("/lib$","");
#else
	path=path.pathname().pathname().pathname();
#endif


	return path;
#endif

	return "";
}

bool System::setCurrentWorkingDirectory(const String &a_path)
{
	bool rv = false;
#if FE_OS==FE_LINUX || FE_OS==FE_OSX
		rv = (0 == chdir(a_path.c_str()));
#elif FE_OS==FE_WIN32 || FE_OS==FE_WIN64
		rv = SetCurrentDirectory(a_path.c_str());
#else
#error
#endif
	return rv;
}

bool System::createDirectory(const String &a_path)
{
	bool rv = false;
#if FE_OS==FE_LINUX || FE_OS==FE_OSX
		rv = !mkdir(a_path.c_str(),0777);
#elif FE_OS==FE_WIN32 || FE_OS==FE_WIN64
		rv = !_mkdir(a_path.c_str());
#else
#error
#endif
	return rv;
}

#if FE_COMPILER==FE_GNU || FE_COMPILER==FE_INTEL
#include <cxxabi.h>

String System::demangle(const String a_symbol,BWORD a_emptyDefault)
{
	char* buffer=NULL;
	FE_UWORD length=0;
	I32 status = 0;

	char* demangled=
			abi::__cxa_demangle(a_symbol.c_str(),buffer,&length,&status);

	String result=demangled? demangled: (a_emptyDefault? "": a_symbol);

	free(demangled);

	return result;
}

#else

String System::demangle(const String a_symbol,BWORD a_emptyDefault)
{
	return a_symbol;
}

#endif

} /* namespace */

